from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor, as_completed
from hashlib import md5
from pathlib import Path
from typing import Any, Dict, List, Optional

import autoflake  # type: ignore
from pyupgrade._data import Settings as PyUpgradeSettings  # type: ignore
from pyupgrade._main import _fix_plugins  # type: ignore

from auto_pytabs.types import VersionTuple, VersionedCode


class Cache:
    instance: Optional[Cache] = None

    def __new__(cls) -> Cache:
        if cls.instance is None:
            cls.instance = super().__new__(cls)
        return cls.instance

    def __init__(self) -> None:
        self.cache_dir = Path(".auto_pytabs_cache")
        self.cache_dir.mkdir(exist_ok=True)
        self._cache: Dict[str, str] = {}
        self._initialised = False
        self._load_cache()

    def _load_cache(self) -> None:
        if self._initialised:
            return
        cache: Dict[str, str] = {}
        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(file.read_text): file.name
                for file in self.cache_dir.iterdir()
            }
            for future in as_completed(futures):
                cache[futures[future]] = future.result()
        self._cache.update(cache)
        self._initialised = True

    @staticmethod
    def make_cache_key(*parts: Any) -> str:
        return md5("".join(map(str, parts)).encode()).hexdigest()

    def get(self, key: str) -> Optional[str]:
        return self._cache.get(key)

    def set(self, key: str, content: str) -> None:
        self._cache[key] = content
        self.cache_dir.joinpath(key).write_text(content)

    def clear(self) -> None:
        for file in self.cache_dir.iterdir():
            file.unlink()


CACHE = Cache()


def upgrade_code(code: str, min_version: VersionTuple, no_cache: bool = False) -> str:
    cache_key: str | None = None
    if not no_cache:
        cache_key = Cache.make_cache_key(code, min_version)

        if cached_code := CACHE.get(cache_key):
            return cached_code

    upgraded_code = _fix_plugins(
        code,
        settings=PyUpgradeSettings(
            min_version=min_version,
            keep_percent_format=True,
            keep_mock=True,
            keep_runtime_typing=True,
        ),
    )
    if upgraded_code != code:
        code = autoflake.fix_code(upgraded_code, remove_all_unused_imports=True)

    if cache_key:
        CACHE.set(cache_key, code)

    return code


def version_code(
    code: str, version_requirements: List[VersionTuple], no_cache: bool = False
) -> VersionedCode:
    latest_code = code
    versioned_code: VersionedCode = {version_requirements[0]: code}

    for version in version_requirements:
        upgraded_code = upgrade_code(latest_code, version, no_cache=no_cache)
        if upgraded_code != latest_code:
            versioned_code[version] = upgraded_code
            latest_code = upgraded_code

    return versioned_code



from __future__ import annotations

import secrets
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, List

import markdown
from mkdocs.config import Config, config_options
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import File, Files

from auto_pytabs.markdown_ext import (
    PendingTransformation,
    convert_block,
    extract_code_blocks,
)
from auto_pytabs.types import VersionTuple
from auto_pytabs.util import get_version_requirements, parse_version_tuple

if TYPE_CHECKING:
    from pymdownx.snippets import SnippetPreprocessor  # type: ignore


class PluginConfig(Config):  # type: ignore[no-untyped-call]
    min_version = config_options.Type(str, default="3.7")
    max_version = config_options.Type(str, default="3.11")
    tmp_path = config_options.Type(Path, default=Path(".autopytabs_tmp"))
    tab_title_template = config_options.Type(str, default="Python {min_version}+")
    no_cache = config_options.Type(bool, default=False)


class AutoPyTabsPlugin(BasePlugin[PluginConfig]):  # type: ignore[no-untyped-call]
    def __init__(self) -> None:
        self.versions: List[VersionTuple] = []
        self.snippets_processor: SnippetPreprocessor | None = None

    def on_config(self, config: MkDocsConfig) -> Config | None:
        min_version = parse_version_tuple(self.config.min_version)
        max_version = parse_version_tuple(self.config.max_version)
        self.versions = get_version_requirements(min_version, max_version)

        if "pymdownx.snippets" in config.markdown_extensions:
            md = markdown.Markdown(
                extensions=config["markdown_extensions"],
                extension_configs=config["mdx_configs"] or {},
            )
            self.snippets_processor = md.preprocessors["snippet"]
        return None

    def _convert_block(self, block: List[str]) -> str:
        return convert_block(
            block=block,
            versions=self.versions,
            tab_title_template=self.config.tab_title_template,
            no_cache=self.config.no_cache,
        )

    def _transform_pending(
        self,
        transformation: PendingTransformation,
    ) -> None:
        new_lines = transformation.new_lines
        to_transform = transformation.to_upgrade

        to_replace = {}
        for index, block in to_transform.items():
            to_replace[index] = self._convert_block(block)

        output = ""
        for i, line in enumerate(new_lines):
            line = to_replace.get(i, line)
            output += line + "\n"
        transformation.tmp_docs_file.write_text(output)

    def on_files(self, files: Files, *, config: MkDocsConfig) -> Files:
        self.config.tmp_path.mkdir(exist_ok=True)

        for file in files:
            if not file.is_documentation_page():
                continue
            if pending := extract_blocks_from_file(
                file,
                self.config.tmp_path,
                self.snippets_processor,
            ):
                self._transform_pending(pending)

        return files

    def _cleanup_temp_files(self) -> None:
        """Cleanup temporary files."""
        if self.config.tmp_path.exists():
            shutil.rmtree(self.config.tmp_path)

    def on_post_build(self, config: MkDocsConfig) -> None:
        self._cleanup_temp_files()

    def on_build_error(self, error: Exception) -> None:
        self._cleanup_temp_files()


def extract_blocks_from_file(
    docs_file: File,
    tmp_path: Path,
    snippets_preprocessor: SnippetPreprocessor | None = None,
) -> PendingTransformation | None:
    content = Path(docs_file.abs_src_path).read_text().splitlines()
    if snippets_preprocessor:
        content = snippets_preprocessor.run(content)

    new_lines, to_upgrade = extract_code_blocks(content)
    if not to_upgrade:
        return None

    tmp_docs_file = tmp_path / secrets.token_hex()
    docs_file.abs_src_path = str(tmp_docs_file)

    return PendingTransformation(
        tmp_docs_file=tmp_docs_file,
        new_lines=new_lines,
        to_upgrade=to_upgrade,
    )



from __future__ import annotations

from typing import List

from auto_pytabs.types import VersionTuple


def get_version_requirements(
    min_version: VersionTuple, max_version: VersionTuple
) -> List[VersionTuple]:
    min_major, min_minor = min_version
    max_major, max_minor = max_version
    return [
        (major, minor)
        for major in range(min_major, max_major + 1)
        for minor in range(min_minor, max_minor + 1)
    ]


def parse_version_tuple(version: str) -> VersionTuple:
    major, minor = version.split(".")
    return int(major), int(minor)


def parse_version_requirements(
    min_version: str, max_version: str
) -> List[VersionTuple]:
    return get_version_requirements(
        parse_version_tuple(min_version), parse_version_tuple(max_version)
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['auto_pytabs']

package_data = \
{'': ['*']}

install_requires = \
['autoflake>=2.0.0,<3.0.0', 'pyupgrade>=3.3.1,<4.0.0']

extras_require = \
{'markdown': ['markdown>=3.2.1,<3.4'],
 'mkdocs': ['mkdocs>=1.4.2,<2.0.0'],
 'sphinx': ['sphinx>=4,<6']}

entry_points = \
{'mkdocs.plugins': ['auto_pytabs = auto_pytabs.mkdocs_plugin:AutoPyTabsPlugin']}

setup_kwargs = {
    'name': 'auto-pytabs',
    'version': '0.1.1',
    'description': 'Automatically generate tabbed code blocks for multiple Python versions in mkdocs, markdown and Sphinx',
    'long_description': '# AutoPyTabs\n\nTooling to automatically generate tabbed code examples for different Python versions in\n[mkdocs](https://www.mkdocs.org) or [Sphinx](https://www.sphinx-doc.org) based documentations, or a plain\n[markdown](https://python-markdown.github.io/) workflow, making use of the\n[pymdown "tabbed"](https://facelessuser.github.io/pymdown-extensions/extensions/tabbed/) markdown extension for markdown,\nand [sphinx{design} tabs](https://sphinx-design.readthedocs.io/en/latest/tabs.html) for Sphinx.\n\n# Work in progress\n\nThis library is work in progress. Approach with caution\n\n## Table of contents\n\n1. [Mkdocs / markdown](#mkdocs-markdown)\n   1. [Plugin or extension?](#plugin-or-extension)\n   1. [Configuration](#mkdocs-config)\n   1. [Examples](#mkdocs-examples)\n   1. [Selectively disable](#selectively-disable)\n   1. [Compatibility with `pymdownx.snippets`](#compatibility-with-pymdownxsnippets)\n1. [Sphinx](#sphinx)\n   1. [Configuration](#sphinx-config)\n   1. [Directives](#directives)\n   1. [Examples](#sphinx-examples)\n   1. [Compatibility with other extensions](#compatibility-with-other-extensions)\n\n# Usage\n\n<h2 id="mkdocs-markdown">Mkdocs / markdown</h2>\n\n### Installation\n\n`pip install auto-pytabs[mkdocs]` or `pip install auto-pytabs[markdown]` if only using\nthe markdown extension.\n\n### Plugin or extension?\n\nWhen using mkdocs, you have the choice to either use AutoPyTabs as a plugin, or a markdown extension.\nThe result is the same, with the main difference being how they integrate into the rendering workflow.\nBehind the scenes, AutoPyTabs makes use of [pyupgrade](https://github.com/asottile/pyupgrade) and\n[autoflake](https://github.com/PyCQA/autoflake) to generate the upgraded code, which is a relatively costly process.\nWhile AutoPyTabs tries to cache as much as possible, due to the limitations of mkdocs, it can still result in a considerable\namount of additional build time.\n\nThe main issues are that the core process of mkdocs is inherently linear and only supports limited incremental builds. To\nsomewhat alleviate these issues, the mkdocs plugin provided by AutPyTabs employs parallelized source-file level transformations:\n\n- After mkdocs has collected all source files, collect the ones that contain code blocks which need upgrading\n- Create temporary files for all collected files and tell mkdocs to use those instead\n- Transform the temporary files, using multiple worker processes\n\n<h3 id="mkdocs-config">Configuration</h3>\n\n**Mkdocs plugin**\n\n```yaml\nsite_name: My Docs\nmarkdown_extensions:\n  - pymdownx.tabbed:\nplugins:\n  - auto_pytabs:\n      min_version: "3.7"  # optional\n      max_version: "3.11" # optional\n      tab_title_template: "Python {min_version}+"  # optional\n      no_cache: false  # optional\n```\n\nOr using the markdown extension directly:\n\n**Markdown extension in mkdocs**\n\n```yaml\nsite_name: My Docs\nmarkdown_extensions:\n  - pymdownx.tabbed:\n  - auto_pytabs:\n      min_version: "3.7"  # optional\n      max_version: "3.11" # optional\n      tab_title_template: "Python {min_version}+"  # optional\n      no_cache: false  # optional\n```\n\n**Markdown extension**\n\n```python\nimport markdown\n\nmd = markdown.Markdown(\n    extensions=["auto_pytabs.markdown_ext"],\n    extension_configs={\n        "auto_pytabs": {\n            "min_version": "3.7",  # optional\n            "max_version": "3.11",  # optional\n            "tab_title_template": "Python {min_version}+",  # optional\n            "no_cache": False,  # optional\n        }\n    },\n)\n\n```\n\n<h3 id="mkdocs-examples">Examples</h3>\n\n**Input**\n\n<pre>\n```python\nfrom typing import Optional, Dict\n\ndef foo(bar: Optional[str]) -> Dict[str, str]:\n    ...\n```\n</pre>\n\n**Equivalent markdown**\n\n<pre>\n=== "Python 3.7+"\n    ```python\n    from typing import Optional, Dict\n\n    def foo(bar: Optional[str]) -> Dict[str, str]:\n        ...\n    ```\n\n=== "Python 3.9+"\n    ```python\n    from typing import Optional\n    \n    \n    def foo(bar: Optional[str]) -> dict[str, str]:\n        ...\n    ```\n\n==== "Python 3.10+"\n    ```python\n    def foo(bar: str | None) -> dict[str, str]:\n        ...\n    ```\n</pre>\n\n#### Nested blocks\n\nNested tabs are supported as well:\n\n**Input**\n\n<pre>\n=== "Level 1-1"\n\n    === "Level 2-1"\n\n        ```python\n        from typing import List\n        x: List[str]\n        ```\n\n    === "Level 2-2"\n    \n        Hello, world!\n\n=== "Level 1-2"\n\n    Goodbye, world!\n</pre>\n\n**Equivalent markdown**\n\n<pre>\n=== "Level 1-1"\n\n    === "Level 2-1"\n\n        === "Python 3.7+"\n            ```python\n            from typing import List\n            x: List[str]\n            ```\n        \n        === "Python 3.9+"\n            ```python\n            x: list[str]\n            ```\n\n    === "Level 2-2"\n\n        Goodbye, world!\n\n=== "Level 1-2"\n    Hello, world!\n    \n</pre>\n\n### Selectively disable\n\nYou can disable conversion for a single code block:\n\n````\n<!-- autopytabs: disable-block -->\n```python\nfrom typing import Set, Optional\n\ndef bar(baz: Optional[str]) -> Set[str]:\n    ...\n```\n````\n\nOr for whole sections / files\n\n```\n<!-- autopytabs: disable -->\neverything after this will be ignored\n<!-- autopytabs: enable -->\nre-enables conversion again\n```\n\n### Compatibility with `pymdownx.snippets`\n\nIf the `pymdownx.snippets` extension is used, make sure that it runs **before** AutoPyTab\n\n## Sphinx\n\nAutPyTabs provides a Sphinx extension `auto_pytabs.sphinx_ext`, enabling its functionality\nfor the `.. code-block` and `.. literalinclude` directives.\n\n### Installation\n\n`pip install auto-pytabs[sphinx]`\n\n### Usage\n\n<h4 id="sphinx-config">Configuration</h4>\n\n```python\nextensions = ["auto_pytabs.sphinx_ext", "sphinx_design"]\n\nauto_pytabs_min_version = (3, 7)  # optional\nauto_pytabs_max_version = (3, 11)  # optional\nauto_pytabs_tab_title_template = "Python {min_version}+"  # optional \nauto_pytabs_no_cache = True  # disabled caching\n```\n\n<h3 id="sphinx-examples">Examples</h3>\n\n**Input**\n\n```rst\n.. code-block:: python\n\n   from typing import Optional, Dict\n   \n   def foo(bar: Optional[str]) -> Dict[str, str]:\n       ...\n```\n\n**Equivalent ReST**\n\n```rst\n.. tab-set::\n\n   .. tab-item:: Python 3.7+\n   \n       .. code-block:: python\n       \n          from typing import Optional, Dict\n      \n          def foo(bar: Optional[str]) -> Dict[str, str]:\n              ...\n\n   .. tab-item:: Python 3.9+\n   \n      .. code-block:: python\n      \n          from typing import Optional\n          \n          \n          def foo(bar: Optional[str]) -> dict[str, str]:\n              ...\n\n   .. tab-item:: Python 3.10+\n   \n      .. code-block:: python\n      \n          def foo(bar: str | None) -> dict[str, str]:\n              ...\n\n```\n\n### Directives\n\nAutoPyTabs overrides the built-in `code-block` and `literal-include` directives,\nextending them with auto-upgrade and tabbing functionality, which means no special\ndirectives, and therefore changes to existing documents are needed.\n\nAdditionally, a `:no-upgrade:` option is added to the directives, which can be used to\nselectively fall back the default behaviour.\n\nTwo new directives are provided as well:\n\n- `.. pytabs-code-block::`\n- `.. pytabs-literalinclude::`\n\nwhich by default act exactly like `.. code-block` and `.. literalinclude` respectively,\nand are mainly to provide AutoPyTab\'s functionality in [compatibility mode](#compatibility-mode).\n\n### Compatibility mode\n\nIf you don\'t want the default behaviour of directive overrides, and instead wish to use the\n`.. pytabs-` directives manually (e.g. because of compatibility issues with other extensions\nor because you only want to apply it to select code blocks) you can make use AutoPyTabs\' compatibility\nmode. To enable it, simply use the `auto_pytabs.sphinx_ext_compat` extension instead of\n`auto_pytabs.sphinx_ext`. Now, only content within `.. pytabs-` directives will be upgraded.\n\n### Compatibility with other extensions\n\nNormally the directive overrides don\'t cause any problems and are very convenient,\nsince no changes to existing documents have to be made. However, if other extensions are included,\nwhich themselves override one of those directives, one of them will inadvertently override the other,\ndepending on the order they\'re defined in `extensions`.\n\nTo combat this, you can use the [compatibility mode](#compatibility-mode) extension instead, which\nonly includes the new directives.\n\nIf you control the conflicting overrides, you can alternatively inherit from\n`auto_py_tabs.sphinx_ext.CodeBlockOverride` and `auto_py_tabs.sphinx_ext.LiteralIncludeOverride`\ninstead of `sphinx.directives.code.CodeBlock` and `sphinx.directives.code.LiteralInclude` respectively.\n',
    'author': 'Janek Nouvertné',
    'author_email': 'provinzkraut@posteo.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import Any, Iterable, TYPE_CHECKING, cast, List, Dict

from docutils.nodes import Node, section, container
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx.config import Config
from sphinx.directives.code import CodeBlock, LiteralInclude
from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import nested_parse_with_titles

from auto_pytabs.core import version_code
from auto_pytabs.util import get_version_requirements

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from auto_pytabs.types import VersionedCode


def indent(string: str, indent_char: str = " ", level: int = 4) -> List[str]:
    return list((indent_char * level) + line for line in string.splitlines())


class UpgradeMixin(SphinxDirective):
    compat: bool = False

    def _render_directive_options(self) -> str:
        ret = ""
        options: Dict[str, Any] = {
            k: v for k, v in self.options.items() if k in CodeBlock.option_spec
        }
        if not self.compat:
            options["no-upgrade"] = True
        for option, value in options.items():
            if self.option_spec[option] is directives.flag:
                value = None
            if isinstance(value, Iterable) and not isinstance(value, str):
                value = "\n".join(value)
            ret += f":{option}: {value if value is not None else ''}\n"
        return ret

    def _create_tabs(
        self,
        versioned_code: VersionedCode,
        tab_title_template: str,
    ) -> List[str]:
        if len(versioned_code) == 1:
            return [
                ".. code-block:: python",
                *indent(self._render_directive_options()),
                "",
                *indent(versioned_code.popitem()[1]),
                "",
            ]

        out = [".. tab-set::", ""]
        for version, code in versioned_code.items():
            version_string = f"{version[0]}.{version[1]}"
            out.extend(
                [
                    f"    .. tab-item:: {tab_title_template.format(min_version=version_string)}",  # noqa: E501
                    f"        :sync: {version_string}",
                    "",
                    "        .. code-block:: python",
                    *indent(self._render_directive_options(), level=12),
                    "",
                    *indent(code, level=12),
                    "",
                ]
            )
        return out

    def _create_py_tab_nodes(self, code: str) -> List[Node]:
        version_requirements = self.config["auto_pytabs_versions"]
        versioned_code = version_code(
            code, version_requirements, no_cache=self.config["auto_pytabs_no_cache"]
        )
        tabs = self._create_tabs(
            versioned_code, self.env.config["auto_pytabs_tab_title_template"]
        )

        rst = ViewList()
        source, lineno = self.get_source_info()
        for line in tabs:
            rst.append(line, source, lineno)

        node = section()
        node.document = self.state.document

        nested_parse_with_titles(self.state, rst, node)
        nodes = node.children

        return cast(List[Node], nodes)


class PyTabsCodeBlock(CodeBlock, UpgradeMixin):
    compat = True

    def run(self) -> list[Node]:
        if self.arguments[0] != "python":
            return super().run()

        return self._create_py_tab_nodes("\n".join(self.content))


class PyTabsLiteralInclude(LiteralInclude, UpgradeMixin):
    compat = True

    def run(self) -> List[Node]:
        base_node = super().run()[0]
        if self.options.get("language") != "python":
            return [base_node]
        if isinstance(base_node, container):
            base_node = base_node.children[1]
        return self._create_py_tab_nodes(base_node.rawsource)  # type: ignore[attr-defined]  # noqa: E501


class CodeBlockOverride(PyTabsCodeBlock):
    compat = False
    option_spec = {**CodeBlock.option_spec, "no-upgrade": directives.flag}

    def run(self) -> List[Node]:
        if "no-upgrade" in self.options:
            return CodeBlock.run(self)

        return super().run()


class LiteralIncludeOverride(PyTabsLiteralInclude):
    compat = False
    option_spec = {**LiteralInclude.option_spec, "no-upgrade": directives.flag}

    def run(self) -> List[Node]:
        if "no-upgrade" in self.options:
            return LiteralInclude.run(self)
        return super().run()


def on_config_inited(app: Sphinx, config: Config) -> None:
    config["auto_pytabs_versions"] = get_version_requirements(
        config["auto_pytabs_min_version"], config["auto_pytabs_max_version"]
    )


def base_setup(app: Sphinx) -> Dict[str, bool | str]:
    app.add_directive("pytabs-code-block", PyTabsCodeBlock)
    app.add_directive("pytabs-literalinclude", PyTabsLiteralInclude)

    app.add_config_value(
        "auto_pytabs_tab_title_template",
        default="Python {min_version}+",
        rebuild="html",
    )
    app.add_config_value("auto_pytabs_min_version", default=(3, 7), rebuild="html")
    app.add_config_value("auto_pytabs_max_version", default=(3, 11), rebuild="html")
    app.add_config_value("auto_pytabs_no_cache", default=False, rebuild="html")

    app.connect("config-inited", on_config_inited)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def setup(app: Sphinx) -> Dict[str, bool | str]:
    app.add_directive("code-block", CodeBlockOverride, override=True)
    app.add_directive("literalinclude", LiteralIncludeOverride, override=True)

    return base_setup(app)



from __future__ import annotations

import re
from pathlib import Path
from typing import Any, Literal, NamedTuple, cast, Set, List, Dict, Tuple

import markdown
from markdown import Extension
from markdown.preprocessors import Preprocessor

from auto_pytabs.core import version_code
from auto_pytabs.util import parse_version_requirements
from auto_pytabs.types import VersionTuple, VersionedCode


RGX_BLOCK_TOKENS = re.compile(r"(.*```py[\w\W]*)|(.*```)")
RGX_PYTABS_DIRECTIVE = re.compile(r"<!-- ?autopytabs: ?(.*)-->")

PyTabDirective = Literal["disable", "enable", "disable-block"]
PYTAB_DIRECTIVES: Set[PyTabDirective] = {"disable", "enable", "disable-block"}


class PendingTransformation(NamedTuple):
    tmp_docs_file: Path
    new_lines: List[str]
    to_upgrade: Dict[int, List[str]]


def _strip_indentation(lines: List[str]) -> Tuple[List[str], str]:
    if not lines:
        return [], ""
    first_line = lines[0]
    indent_char = ""
    if first_line[0] in [" ", "\t"]:
        indent_char = first_line[0]
    indent = indent_char * (len(first_line) - len(first_line.lstrip(indent_char)))
    if indent:
        return [line.split(indent, 1)[1] if line else "" for line in lines], indent
    return lines, indent


def _add_indentation(code: str, indentation: str) -> str:
    return "\n".join(indentation + line for line in code.splitlines())


def _get_pytabs_directive(line: str) -> PyTabDirective | None:
    match = RGX_PYTABS_DIRECTIVE.match(line)
    if match:
        matched_directive = match.group(1).strip()
        if matched_directive in PYTAB_DIRECTIVES:
            return cast(PyTabDirective, matched_directive)
        raise RuntimeError(f"Invalid AutoPytabs directive: {matched_directive!r}")
    return None


def extract_code_blocks(lines: List[str]) -> Tuple[List[str], Dict[int, List[str]]]:
    in_block = False
    enabled = True
    new_lines: List[str] = []

    to_transform = {}

    start = 0
    for i, line in enumerate(lines):
        is_comment_line = False
        directive = _get_pytabs_directive(line)
        if directive:
            is_comment_line = True
            if directive == "disable":
                enabled = False
            elif directive == "enable":
                enabled = True

        match = RGX_BLOCK_TOKENS.match(line)
        if match:
            if match.group(1):
                in_block = True
                start = i
            elif match.group(2) and in_block:
                in_block = False
                block = lines[start : i + 1]
                block_directive = _get_pytabs_directive(lines[start - 1])
                if enabled and block_directive != "disable-block":
                    to_transform[len(new_lines)] = block
                    new_lines.append("")
                else:
                    new_lines.extend(block)
            else:
                new_lines.append(line)
        elif not in_block and not is_comment_line:
            new_lines.append(line)

    return new_lines, to_transform


def _build_tabs(
    *, versioned_code: VersionedCode, head: str, tail: str, tab_title_template: str
) -> str:
    out = []
    for version, code in versioned_code.items():
        version_string = f"{version[0]}.{version[1]}"
        lines = [head, *code.splitlines(), tail]
        code = "\n".join("    " + line for line in lines)
        tab_title = tab_title_template.format(min_version=version_string)
        out.append(f'=== "{tab_title}"\n{code}\n')
    return "\n".join(out)


def convert_block(
    *,
    block: List[str],
    versions: List[VersionTuple],
    tab_title_template: str,
    no_cache: bool,
) -> str:
    block, indentation = _strip_indentation(block)
    head, *code_lines, tail = block
    code = "\n".join(code_lines)
    versioned_code = version_code(code, versions, no_cache=no_cache)

    if len(versioned_code) > 1:
        code = _build_tabs(
            versioned_code=versioned_code,
            head=head,
            tail=tail,
            tab_title_template=tab_title_template,
        )
    else:
        code = "\n".join([head, versioned_code[versions[0]], tail])

    code = _add_indentation(code, indentation)

    return code


class UpgradePreprocessor(Preprocessor):
    def __init__(
        self,
        *args: Any,
        min_version: str,
        max_version: str,
        tab_title_template: str | None = None,
        no_cache: bool = False,
        **kwargs: Any,
    ) -> None:
        self.versions = parse_version_requirements(min_version, max_version)
        self.tab_title_template = tab_title_template or "Python {min_version}+"
        self.no_cache = no_cache
        super().__init__(*args, **kwargs)

    def run(self, lines: List[str]) -> List[str]:
        new_lines, to_transform = extract_code_blocks(lines)

        output_lines = []
        for i, line in enumerate(new_lines):
            block_to_transform = to_transform.get(i)
            if block_to_transform:
                transformed_block = convert_block(
                    block=block_to_transform,
                    versions=self.versions,
                    tab_title_template=self.tab_title_template,
                    no_cache=self.no_cache,
                ).splitlines()
                output_lines.extend(transformed_block)
            else:
                output_lines.append(line)

        return output_lines


class AutoPyTabsExtension(Extension):
    def __init__(self, *args: Any, **kwargs: Any):
        """Initialize."""

        self.config = {
            "min_version": ["3.7", "minimum version"],
            "max_version": ["3.11", "maximum version"],
            "tab_title_template": ["", "tab title format-string"],
            "no_cache": [False, "disable caching"],
        }
        super().__init__(*args, **kwargs)

    def extendMarkdown(self, md: markdown.Markdown) -> None:
        """Register the extension."""

        self.md = md
        md.registerExtension(self)
        config = self.getConfigs()
        md.preprocessors.register(
            UpgradePreprocessor(
                min_version=config["min_version"],
                max_version=config["max_version"],
                tab_title_template=config["tab_title_template"],
            ),
            "auto_pytabs",
            32,
        )


def makeExtension(**kwargs: Any) -> AutoPyTabsExtension:
    return AutoPyTabsExtension(**kwargs)
